@@ -5,6 +5,7 @@
  * Copyright (c) 2004-2005 Richard Russon
  * Copyright (c) 2005-2006 Yura Pakhuchiy
  * Copyright (c) 2005-2008 Szabolcs Szakacsits
+ * Copyright (c) 2007-2021 Jean-Pierre Andre
  *
  * This program/include file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as published
@@ -37,13 +38,14 @@
 #endif
 
 #include "attrib.h"
-#include "collate.h"
 #include "debug.h"
 #include "index.h"
+#include "collate.h"
 #include "mst.h"
 #include "dir.h"
 #include "logging.h"
 #include "bitmap.h"
+#include "reparse.h"
 #include "misc.h"
 
 /**
@@ -82,7 +84,7 @@ static int ntfs_ib_write(ntfs_index_context *icx, INDEX_BLOCK *ib)
 {
 	s64 ret, vcn = sle64_to_cpu(ib->index_block_vcn);
 	
-	ntfs_log_trace("vcn: %lld\n", vcn);
+	ntfs_log_trace("vcn: %lld\n", (long long)vcn);
 	
 	ret = ntfs_attr_mst_pwrite(icx->ia_na, ntfs_ib_vcn_to_pos(icx, vcn),
 				   1, icx->block_size, ib);
@@ -141,7 +143,7 @@ static void ntfs_index_ctx_free(ntfs_index_context *icx)
 {
 	ntfs_log_trace("Entering\n");
 	
-	if (!icx->entry)
+	if (!icx->bad_index && !icx->entry)
 		return;
 
 	if (icx->actx)
@@ -189,9 +191,9 @@ void ntfs_index_ctx_reinit(ntfs_index_context *icx)
 	};
 }
 
-static VCN *ntfs_ie_get_vcn_addr(INDEX_ENTRY *ie)
+static leVCN *ntfs_ie_get_vcn_addr(INDEX_ENTRY *ie)
 {
-	return (VCN *)((u8 *)ie + le16_to_cpu(ie->length) - sizeof(VCN));
+	return (leVCN *)((u8 *)ie + le16_to_cpu(ie->length) - sizeof(leVCN));
 }
 
 /**
@@ -336,7 +338,7 @@ static void ntfs_ie_delete(INDEX_HEADER *ih, INDEX_ENTRY *ie)
 
 static void ntfs_ie_set_vcn(INDEX_ENTRY *ie, VCN vcn)
 {
-	*ntfs_ie_get_vcn_addr(ie) = cpu_to_le64(vcn);
+	*ntfs_ie_get_vcn_addr(ie) = cpu_to_sle64(vcn);
 }
 
 /**
@@ -386,42 +388,6 @@ static INDEX_ENTRY *ntfs_ie_dup_novcn(INDEX_ENTRY *ie)
 	return dup;
 }
 
-static int ntfs_ia_check(ntfs_index_context *icx, INDEX_BLOCK *ib, VCN vcn)
-{
-	u32 ib_size = (unsigned)le32_to_cpu(ib->index.allocated_size) + 0x18;
-	
-	ntfs_log_trace("Entering\n");
-	
-	if (!ntfs_is_indx_record(ib->magic)) {
-		
-		ntfs_log_error("Corrupt index block signature: vcn %lld inode "
-			       "%llu\n", (long long)vcn,
-			       (unsigned long long)icx->ni->mft_no);
-		return -1;
-	}
-	
-	if (sle64_to_cpu(ib->index_block_vcn) != vcn) {
-		
-		ntfs_log_error("Corrupt index block: VCN (%lld) is different "
-			       "from expected VCN (%lld) in inode %llu\n",
-			       (long long)sle64_to_cpu(ib->index_block_vcn),
-			       (long long)vcn,
-			       (unsigned long long)icx->ni->mft_no);
-		return -1;
-	}
-	
-	if (ib_size != icx->block_size) {
-		
-		ntfs_log_error("Corrupt index block : VCN (%lld) of inode %llu "
-			       "has a size (%u) differing from the index "
-			       "specified size (%u)\n", (long long)vcn, 
-			       (unsigned long long)icx->ni->mft_no, ib_size,
-			       icx->block_size);
-		return -1;
-	}
-	return 0;
-}
-
 static INDEX_ROOT *ntfs_ir_lookup(ntfs_inode *ni, ntfschar *name,
 				  u32 name_len, ntfs_attr_search_ctx **ctx)
 {
@@ -467,6 +433,133 @@ static INDEX_ROOT *ntfs_ir_lookup2(ntfs_inode *ni, ntfschar *name, u32 len)
 	return ir;
 }
 
+/*
+ *		Check the consistency of an index block
+ *
+ *	Make sure the index block does not overflow from the index record.
+ *	The size of block is assumed to have been checked to be what is
+ *	defined in the index root.
+ *
+ *	Returns 0 if no error was found
+ *		-1 otherwise (with errno unchanged)
+ *
+ *      |<--->|  offsetof(INDEX_BLOCK, index)
+ *      |     |<--->|  sizeof(INDEX_HEADER)
+ *      |     |     |
+ *      |     |     | seq          index entries         unused
+ *      |=====|=====|=====|===========================|==============|
+ *      |     |           |                           |              |
+ *      |     |<--------->| entries_offset            |              |
+ *      |     |<---------------- index_length ------->|              |
+ *      |     |<--------------------- allocated_size --------------->|
+ *      |<--------------------------- block_size ------------------->|
+ *
+ *      size(INDEX_HEADER) <= ent_offset < ind_length <= alloc_size < bk_size
+ */
+
+int ntfs_index_block_inconsistent(const INDEX_BLOCK *ib, u32 block_size,
+			u64 inum, VCN vcn)
+{
+	u32 ib_size = (unsigned)le32_to_cpu(ib->index.allocated_size)
+			+ offsetof(INDEX_BLOCK, index);
+
+	if (!ntfs_is_indx_record(ib->magic)) {
+		ntfs_log_error("Corrupt index block signature: vcn %lld inode "
+			       "%llu\n", (long long)vcn,
+			       (unsigned long long)inum);
+		return -1;
+	}
+	
+	if (sle64_to_cpu(ib->index_block_vcn) != vcn) {
+		ntfs_log_error("Corrupt index block: VCN (%lld) is different "
+			       "from expected VCN (%lld) in inode %llu\n",
+			       (long long)sle64_to_cpu(ib->index_block_vcn),
+			       (long long)vcn,
+			       (unsigned long long)inum);
+		return -1;
+	}
+	
+	if (ib_size != block_size) {
+		ntfs_log_error("Corrupt index block : VCN (%lld) of inode %llu "
+			       "has a size (%u) differing from the index "
+			       "specified size (%u)\n", (long long)vcn, 
+			       (unsigned long long)inum, ib_size,
+			       (unsigned int)block_size);
+		return -1;
+	}
+	if (le32_to_cpu(ib->index.entries_offset) < sizeof(INDEX_HEADER)) {
+		ntfs_log_error("Invalid index entry offset in inode %lld\n",
+				(unsigned long long)inum);
+		return -1;
+	}
+	if (le32_to_cpu(ib->index.index_length)
+			<= le32_to_cpu(ib->index.entries_offset)) {
+		ntfs_log_error("No space for index entries in inode %lld\n",
+				(unsigned long long)inum);
+		return -1;
+	}
+	if (le32_to_cpu(ib->index.allocated_size)
+			< le32_to_cpu(ib->index.index_length)) {
+		ntfs_log_error("Index entries overflow in inode %lld\n",
+				(unsigned long long)inum);
+		return -1;
+	}
+
+	return (0);
+}
+
+
+/*
+ *		Check the consistency of an index entry
+ *
+ *	Make sure data and key do not overflow from entry.
+ *	As a side effect, an entry with zero length is rejected.
+ *	This entry must be a full one (no INDEX_ENTRY_END flag), and its
+ *	length must have been checked beforehand to not overflow from the
+ *	index record.
+ *
+ *	Returns 0 if no error was found
+ *		-1 otherwise (with errno unchanged)
+ */
+
+int ntfs_index_entry_inconsistent(const INDEX_ENTRY *ie,
+			COLLATION_RULES collation_rule, u64 inum)
+{
+	int ret;
+
+	ret = 0;
+	if (ie->key_length
+		&& ((le16_to_cpu(ie->key_length) + offsetof(INDEX_ENTRY, key))
+			    > le16_to_cpu(ie->length))) {
+		ntfs_log_error("Overflow from index entry in inode %lld\n",
+				(long long)inum);
+		ret = -1;
+
+	} else
+		if (collation_rule == COLLATION_FILE_NAME) {
+			if ((offsetof(INDEX_ENTRY, key.file_name.file_name)
+				    + ie->key.file_name.file_name_length
+						* sizeof(ntfschar))
+				> le16_to_cpu(ie->length)) {
+				ntfs_log_error("File name overflow from index"
+					" entry in inode %lld\n",
+					(long long)inum);
+				ret = -1;
+			}
+		} else {
+			if (ie->data_length
+				&& ((le16_to_cpu(ie->data_offset)
+				    + le16_to_cpu(ie->data_length))
+				    > le16_to_cpu(ie->length))) {
+				ntfs_log_error("Data overflow from index"
+					" entry in inode %lld\n",
+					(long long)inum);
+				ret = -1;
+			}
+		}
+	return (ret);
+}
+
 /** 
  * Find a key in the index block.
  * 
@@ -515,8 +608,19 @@ static int ntfs_ie_lookup(const void *key, const int key_len,
 		 * Not a perfect match, need to do full blown collation so we
 		 * know which way in the B+tree we have to go.
 		 */
-		rc = ntfs_collate(icx->ni->vol, icx->cr, key, key_len, &ie->key,
-				  le16_to_cpu(ie->key_length));
+		if (!icx->collate) {
+			ntfs_log_error("Collation function not defined\n");
+			errno = EOPNOTSUPP;
+			return STATUS_ERROR;
+		}
+			/* Make sure key and data do not overflow from entry */
+		if (ntfs_index_entry_inconsistent(ie, icx->ir->collation_rule,
+				icx->ni->mft_no)) {
+			errno = EIO;
+			return STATUS_ERROR;
+		}
+		rc = icx->collate(icx->ni->vol, key, key_len,
+					&ie->key, le16_to_cpu(ie->key_length));
 		if (rc == NTFS_COLLATION_ERROR) {
 			ntfs_log_error("Collation error. Perhaps a filename "
 				       "contains invalid characters?\n");
@@ -556,7 +660,7 @@ static int ntfs_ie_lookup(const void *key, const int key_len,
 	*vcn = ntfs_ie_get_vcn(ie);
 	if (*vcn < 0) {
 		errno = EINVAL;
-		ntfs_log_perror("Negative vcn in inode %llu\n",
+		ntfs_log_perror("Negative vcn in inode %llu",
 			       	(unsigned long long)icx->ni->mft_no);
 		return STATUS_ERROR;
 	}
@@ -585,7 +689,7 @@ static int ntfs_ib_read(ntfs_index_context *icx, VCN vcn, INDEX_BLOCK *dst)
 {
 	s64 pos, ret;
 
-	ntfs_log_trace("vcn: %lld\n", vcn);
+	ntfs_log_trace("vcn: %lld\n", (long long)vcn);
 	
 	pos = ntfs_ib_vcn_to_pos(icx, vcn);
 
@@ -599,8 +703,11 @@ static int ntfs_ib_read(ntfs_index_context *icx, VCN vcn, INDEX_BLOCK *dst)
 		return -1;
 	}
 	
-	if (ntfs_ia_check(icx, dst, vcn))
+	if (ntfs_index_block_inconsistent((INDEX_BLOCK*)dst, icx->block_size,
+				icx->ia_na->ni->mft_no, vcn)) {
+		errno = EIO;
 		return -1;
+	}
 	
 	return 0;
 }
@@ -694,25 +801,22 @@ int ntfs_index_lookup(const void *key, const int key_len, ntfs_index_context *ic
 	if (ni->vol->cluster_size <= icx->block_size)
 		icx->vcn_size_bits = ni->vol->cluster_size_bits;
 	else
-		icx->vcn_size_bits = ni->vol->sector_size_bits;
-		
-	icx->cr = ir->collation_rule;
-	if (!ntfs_is_collation_rule_supported(icx->cr)) {
+		icx->vcn_size_bits = NTFS_BLOCK_SIZE_BITS;
+			/* get the appropriate collation function */
+	icx->ir = ir;
+	icx->collate = ntfs_get_collate_function(ir->collation_rule);
+	if (!icx->collate) {
 		err = errno = EOPNOTSUPP;
 		ntfs_log_perror("Unknown collation rule 0x%x", 
-				(unsigned)le32_to_cpu(icx->cr));
+				(unsigned)le32_to_cpu(ir->collation_rule));
 		goto err_out;
 	}
 	
 	old_vcn = VCN_INDEX_ROOT_PARENT;
-	/* 
-	 * FIXME: check for both ir and ib that the first index entry is
-	 * within the index block.
-	 */
 	ret = ntfs_ie_lookup(key, key_len, icx, &ir->index, &vcn, &ie);
 	if (ret == STATUS_ERROR) {
 		err = errno;
-		goto err_out;
+		goto err_lookup;
 	}
 	
 	icx->ir = ir;
@@ -746,7 +850,7 @@ descend_into_child_node:
 	}
 	old_vcn = vcn;
 
-	ntfs_log_debug("Descend into node with VCN %lld.\n", vcn);
+	ntfs_log_debug("Descend into node with VCN %lld\n", (long long)vcn);
 	
 	if (ntfs_ib_read(icx, vcn, ib))
 		goto err_out;
@@ -773,6 +877,12 @@ descend_into_child_node:
 	
 	goto descend_into_child_node;
 err_out:
+	icx->bad_index = TRUE;	/* Force icx->* to be freed */
+err_lookup:
+	if (icx->actx) {
+		ntfs_attr_put_search_ctx(icx->actx);
+		icx->actx = NULL;
+	}
 	free(ib);
 	if (!err)
 		err = EIO;
@@ -797,24 +907,24 @@ static INDEX_BLOCK *ntfs_ib_alloc(VCN ib_vcn, u32 ib_size,
 	INDEX_BLOCK *ib;
 	int ih_size = sizeof(INDEX_HEADER);
 	
-	ntfs_log_trace("Entering ib_vcn = %lld ib_size = %u\n", ib_vcn, ib_size);
+	ntfs_log_trace("ib_vcn: %lld ib_size: %u\n", (long long)ib_vcn, ib_size);
 	
 	ib = ntfs_calloc(ib_size);
 	if (!ib)
 		return NULL;
 	
 	ib->magic = magic_INDX;
-	ib->usa_ofs = cpu_to_le16(sizeof(INDEX_BLOCK));
+	ib->usa_ofs = const_cpu_to_le16(sizeof(INDEX_BLOCK));
 	ib->usa_count = cpu_to_le16(ib_size / NTFS_BLOCK_SIZE + 1);
 	/* Set USN to 1 */
-	*(u16 *)((char *)ib + le16_to_cpu(ib->usa_ofs)) = cpu_to_le16(1);
-	ib->lsn = cpu_to_le64(0);
+	*(le16 *)((char *)ib + le16_to_cpu(ib->usa_ofs)) = const_cpu_to_le16(1);
+	ib->lsn = const_cpu_to_sle64(0);
 	
 	ib->index_block_vcn = cpu_to_sle64(ib_vcn);
 	
 	ib->index.entries_offset = cpu_to_le32((ih_size +
 			le16_to_cpu(ib->usa_count) * 2 + 7) & ~7);
-	ib->index.index_length = 0;
+	ib->index.index_length = const_cpu_to_le32(0);
 	ib->index.allocated_size = cpu_to_le32(ib_size - 
 					       (sizeof(INDEX_BLOCK) - ih_size));
 	ib->index.ih_flags = node_type;
@@ -893,7 +1003,7 @@ static int ntfs_ibm_modify(ntfs_index_context *icx, VCN vcn, int set)
 	ntfs_attr *na;
 	int ret = STATUS_ERROR;
 
-	ntfs_log_trace("%s vcn: %lld\n", set ? "set" : "clear", vcn);
+	ntfs_log_trace("%s vcn: %lld\n", set ? "set" : "clear", (long long)vcn);
 	
 	na = ntfs_attr_open(icx->ni, AT_BITMAP,  icx->name, icx->name_len);
 	if (!na) {
@@ -970,7 +1080,7 @@ static VCN ntfs_ibm_get_free(ntfs_index_context *icx)
 	
 	vcn = ntfs_ibm_pos_to_vcn(icx, size * 8);
 out:	
-	ntfs_log_trace("allocated vcn: %lld\n", vcn);
+	ntfs_log_trace("allocated vcn: %lld\n", (long long)vcn);
 
 	if (ntfs_ibm_set(icx, vcn))
 		vcn = (VCN)-1;
@@ -1114,6 +1224,7 @@ static int ntfs_ir_reparent(ntfs_index_context *icx)
 	INDEX_ENTRY *ie;
 	INDEX_BLOCK *ib = NULL;
 	VCN new_ib_vcn;
+	int ix_root_size;
 	int ret = STATUS_ERROR;
 
 	ntfs_log_trace("Entering\n");
@@ -1143,6 +1254,7 @@ static int ntfs_ir_reparent(ntfs_index_context *icx)
 	if (ntfs_ib_write(icx, ib))
 		goto clear_bmp;
 	
+retry :
 	ir = ntfs_ir_lookup(icx->ni, icx->name, icx->name_len, &ctx);
 	if (!ir)
 		goto clear_bmp;
@@ -1151,18 +1263,37 @@ static int ntfs_ir_reparent(ntfs_index_context *icx)
 	
 	ie = ntfs_ie_get_first(&ir->index);
 	ie->ie_flags |= INDEX_ENTRY_NODE;
-	ie->length = cpu_to_le16(sizeof(INDEX_ENTRY_HEADER) + sizeof(VCN));
+	ie->length = const_cpu_to_le16(sizeof(INDEX_ENTRY_HEADER) + sizeof(VCN));
 	
 	ir->index.ih_flags = LARGE_INDEX;
 	ir->index.index_length = cpu_to_le32(le32_to_cpu(ir->index.entries_offset)
 					     + le16_to_cpu(ie->length));
 	ir->index.allocated_size = ir->index.index_length;
-	
+	ix_root_size = sizeof(INDEX_ROOT) - sizeof(INDEX_HEADER)
+			+ le32_to_cpu(ir->index.allocated_size);
 	if (ntfs_resident_attr_value_resize(ctx->mrec, ctx->attr,
-			sizeof(INDEX_ROOT) - sizeof(INDEX_HEADER) +
-			le32_to_cpu(ir->index.allocated_size)))
+					ix_root_size)) {
+			/*
+			 * When there is no space to build a non-resident
+			 * index, we may have to move the root to an extent
+			 */
+		if ((errno == ENOSPC)
+		    && (ctx->al_entry || !ntfs_inode_add_attrlist(icx->ni))) {
+			ntfs_attr_put_search_ctx(ctx);
+			ctx = (ntfs_attr_search_ctx*)NULL;
+			ir = ntfs_ir_lookup(icx->ni, icx->name, icx->name_len,
+							&ctx);
+			if (ir
+			    && !ntfs_attr_record_move_away(ctx, ix_root_size
+				    - le32_to_cpu(ctx->attr->value_length))) {
+				ntfs_attr_put_search_ctx(ctx);
+				ctx = (ntfs_attr_search_ctx*)NULL;
+				goto retry;
+			}
+		}
 		/* FIXME: revert index root */
 		goto clear_bmp;
+	}
 	/*
 	 *  FIXME: do it earlier if we have enough space in IR (should always),
 	 *  so in error case we wouldn't lose the IB.
@@ -1228,16 +1359,16 @@ static int ntfs_ir_make_space(ntfs_index_context *icx, int data_size)
 	int ret;
 
 	ntfs_log_trace("Entering\n");
-	
+
 	ret = ntfs_ir_truncate(icx, data_size);
 	if (ret == STATUS_RESIDENT_ATTRIBUTE_FILLED_MFT) {
-		
+	
 		ret = ntfs_ir_reparent(icx);
 		if (ret == STATUS_OK)
 			ret = STATUS_KEEP_SEARCHING;
 		else
 			ntfs_log_perror("Failed to nodify INDEX_ROOT");
-	} 
+	}
 
 	return ret;
 }
@@ -1418,17 +1549,21 @@ static int ntfs_ib_split(ntfs_index_context *icx, INDEX_BLOCK *ib)
 	return ret;
 }
 
-
-static int ntfs_ie_add(ntfs_index_context *icx, INDEX_ENTRY *ie)
+/* JPA static */
+int ntfs_ie_add(ntfs_index_context *icx, INDEX_ENTRY *ie)
 {
-	char *fn;
 	INDEX_HEADER *ih;
 	int allocated_size, new_size;
 	int ret = STATUS_ERROR;
 	
+#ifdef DEBUG
+/* removed by JPA to make function usable for security indexes
+	char *fn;
 	fn = ntfs_ie_filename_get(ie);
 	ntfs_log_trace("file: '%s'\n", fn);
 	ntfs_attr_name_free(&fn);
+*/
+#endif
 	
 	while (1) {
 				
@@ -1529,19 +1664,32 @@ static int ntfs_ih_takeout(ntfs_index_context *icx, INDEX_HEADER *ih,
 			   INDEX_ENTRY *ie, INDEX_BLOCK *ib)
 {
 	INDEX_ENTRY *ie_roam;
+	int freed_space;
+	BOOL full;
 	int ret = STATUS_ERROR;
 	
 	ntfs_log_trace("Entering\n");
 	
+	full = ih->index_length == ih->allocated_size;
 	ie_roam = ntfs_ie_dup_novcn(ie);
 	if (!ie_roam)
 		return STATUS_ERROR;
 
 	ntfs_ie_delete(ih, ie);
 
-	if (ntfs_icx_parent_vcn(icx) == VCN_INDEX_ROOT_PARENT)
+	if (ntfs_icx_parent_vcn(icx) == VCN_INDEX_ROOT_PARENT) {
+		/*
+		 * Recover the space which may have been freed
+		 * while deleting an entry from root index
+		 */
+		freed_space = le32_to_cpu(ih->allocated_size)
+				- le32_to_cpu(ih->index_length);
+		if (full && (freed_space > 0) && !(freed_space & 7)) {
+			ntfs_ir_truncate(icx, le32_to_cpu(ih->index_length));
+			/* do nothing if truncation fails */
+		}
 		ntfs_inode_mark_dirty(icx->actx->ntfs_ino);
-	else
+	} else
 		if (ntfs_ib_write(icx, ib))
 			goto out;
 	
@@ -1765,7 +1913,8 @@ out:
  *
  * Return 0 on success or -1 on error with errno set to the error code.
  */
-static int ntfs_index_rm(ntfs_index_context *icx)
+/*static JPA*/
+int ntfs_index_rm(ntfs_index_context *icx)
 {
 	INDEX_HEADER *ih;
 	int err, ret = STATUS_OK;
@@ -1808,12 +1957,14 @@ err_out:
 	goto out;
 }
 
-int ntfs_index_remove(ntfs_inode *ni, const void *key, const int keylen)
+int ntfs_index_remove(ntfs_inode *dir_ni,
+		ntfs_inode *ni __attribute__((unused)),
+		const void *key, const int keylen)
 {
 	int ret = STATUS_ERROR;
 	ntfs_index_context *icx;
 
-	icx = ntfs_index_ctx_get(ni, NTFS_INDEX_I30, 4);
+	icx = ntfs_index_ctx_get(dir_ni, NTFS_INDEX_I30, 4);
 	if (!icx)
 		return -1;
 
@@ -1822,12 +1973,6 @@ int ntfs_index_remove(ntfs_inode *ni, const void *key, const int keylen)
 		if (ntfs_index_lookup(key, keylen, icx))
 			goto err_out;
 
-		if (((FILE_NAME_ATTR *)icx->data)->file_attributes &
-				FILE_ATTR_REPARSE_POINT) {
-			errno = EOPNOTSUPP;
-			goto err_out;
-		}
-
 		ret = ntfs_index_rm(icx);
 		if (ret == STATUS_ERROR)
 			goto err_out;
@@ -1883,3 +2028,167 @@ out:
 }
 
 
+/*
+ *		Walk down the index tree (leaf bound)
+ *	until there are no subnode in the first index entry
+ *	returns the entry at the bottom left in subnode
+ */
+
+static INDEX_ENTRY *ntfs_index_walk_down(INDEX_ENTRY *ie,
+			ntfs_index_context *ictx)
+{
+	INDEX_ENTRY *entry;
+	s64 vcn;
+
+	entry = ie;
+	do {
+		vcn = ntfs_ie_get_vcn(entry);
+		if (ictx->is_in_root) {
+
+			/* down from level zero */
+
+			ictx->ir = (INDEX_ROOT*)NULL;
+			ictx->ib = (INDEX_BLOCK*)ntfs_malloc(ictx->block_size);
+			ictx->pindex = 1;
+			ictx->is_in_root = FALSE;
+		} else {
+
+			/* down from non-zero level */
+			
+			ictx->pindex++;
+		}
+		ictx->parent_pos[ictx->pindex] = 0;
+		ictx->parent_vcn[ictx->pindex] = vcn;
+		if (!ntfs_ib_read(ictx,vcn,ictx->ib)) {
+			ictx->entry = ntfs_ie_get_first(&ictx->ib->index);
+			entry = ictx->entry;
+		} else
+			entry = (INDEX_ENTRY*)NULL;
+	} while (entry && (entry->ie_flags & INDEX_ENTRY_NODE));
+	return (entry);
+}
+
+/*
+ *		Walk up the index tree (root bound)
+ *	until there is a valid data entry in parent
+ *	returns the parent entry or NULL if no more parent
+ */
+
+static INDEX_ENTRY *ntfs_index_walk_up(INDEX_ENTRY *ie,
+			ntfs_index_context *ictx)
+{
+	INDEX_ENTRY *entry;
+	s64 vcn;
+
+	entry = ie;
+	if (ictx->pindex > 0) {
+		do {
+			ictx->pindex--;
+			if (!ictx->pindex) {
+
+					/* we have reached the root */
+
+				free(ictx->ib);
+				ictx->ib = (INDEX_BLOCK*)NULL;
+				ictx->is_in_root = TRUE;
+				/* a new search context is to be allocated */
+				if (ictx->actx)
+					free(ictx->actx);
+				ictx->ir = ntfs_ir_lookup(ictx->ni,
+					ictx->name, ictx->name_len,
+					&ictx->actx);
+				if (ictx->ir)
+					entry = ntfs_ie_get_by_pos(
+						&ictx->ir->index,
+						ictx->parent_pos[ictx->pindex]);
+				else
+					entry = (INDEX_ENTRY*)NULL;
+			} else {
+					/* up into non-root node */
+				vcn = ictx->parent_vcn[ictx->pindex];
+				if (!ntfs_ib_read(ictx,vcn,ictx->ib)) {
+					entry = ntfs_ie_get_by_pos(
+						&ictx->ib->index,
+						ictx->parent_pos[ictx->pindex]);
+				} else
+					entry = (INDEX_ENTRY*)NULL;
+			}
+		ictx->entry = entry;
+		} while (entry && (ictx->pindex > 0)
+			 && (entry->ie_flags & INDEX_ENTRY_END));
+	} else
+		entry = (INDEX_ENTRY*)NULL;
+	return (entry);
+}
+
+/*
+ *		Get next entry in an index according to collating sequence.
+ *	Must be initialized through a ntfs_index_lookup()
+ *
+ *	Returns next entry or NULL if none
+ *
+ *	Sample layout :
+ *
+ *                 +---+---+---+---+---+---+---+---+    n ptrs to subnodes
+ *                 |   |   | 10| 25| 33|   |   |   |    n-1 keys in between
+ *                 +---+---+---+---+---+---+---+---+    no key in last entry
+ *                              | A | A
+ *                              | | | +-------------------------------+
+ *   +--------------------------+ | +-----+                           |
+ *   |                            +--+    |                           |
+ *   V                               |    V                           |
+ * +---+---+---+---+---+---+---+---+ |  +---+---+---+---+---+---+---+---+
+ * | 11| 12| 13| 14| 15| 16| 17|   | |  | 26| 27| 28| 29| 30| 31| 32|   |
+ * +---+---+---+---+---+---+---+---+ |  +---+---+---+---+---+---+---+---+
+ *                               |   |
+ *       +-----------------------+   |
+ *       |                           |
+ *     +---+---+---+---+---+---+---+---+
+ *     | 18| 19| 20| 21| 22| 23| 24|   |
+ *     +---+---+---+---+---+---+---+---+
+ */
+
+INDEX_ENTRY *ntfs_index_next(INDEX_ENTRY *ie, ntfs_index_context *ictx)
+{
+	INDEX_ENTRY *next;
+	le16 flags;
+
+			/*
+                         * lookup() may have returned an invalid node
+			 * when searching for a partial key
+			 * if this happens, walk up
+			 */
+
+	if (ie->ie_flags & INDEX_ENTRY_END)
+		next = ntfs_index_walk_up(ie, ictx);
+	else {
+			/*
+			 * get next entry in same node
+                         * there is always one after any entry with data
+			 */
+
+		next = (INDEX_ENTRY*)((char*)ie + le16_to_cpu(ie->length));
+		++ictx->parent_pos[ictx->pindex];
+		flags = next->ie_flags;
+
+			/* walk down if it has a subnode */
+
+		if (flags & INDEX_ENTRY_NODE) {
+			next = ntfs_index_walk_down(next,ictx);
+		} else {
+
+				/* walk up it has no subnode, nor data */
+
+			if (flags & INDEX_ENTRY_END) {
+				next = ntfs_index_walk_up(next, ictx);
+			}
+		}
+	}
+		/* return NULL if stuck at end of a block */
+
+	if (next && (next->ie_flags & INDEX_ENTRY_END))
+		next = (INDEX_ENTRY*)NULL;
+	return (next);
+}
+
+
