@@ -3,7 +3,8 @@
  *
  * Copyright (c) 2002-2004 Anton Altaparmakov
  * Copyright (c) 2004 Yura Pakhuchiy
- * Copyright (c) 2004-2007 Szabolcs Szakacsits
+ * Copyright (c) 2004-2008 Szabolcs Szakacsits
+ * Copyright (c) 2008-2009 Jean-Pierre Andre
  *
  * This program/include file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as published
@@ -54,9 +55,15 @@
 #define NTFS_LCNALLOC_BSIZE 4096
 #define NTFS_LCNALLOC_SKIP  NTFS_LCNALLOC_BSIZE
 
+enum {
+	ZONE_MFT = 1,
+	ZONE_DATA1 = 2,
+	ZONE_DATA2 = 4
+} ;
+
 static void ntfs_cluster_set_zone_pos(LCN start, LCN end, LCN *pos, LCN tc)
 {
-	ntfs_log_trace("pos: %lld  tc: %lld\n", (long long)*pos, tc);
+	ntfs_log_trace("pos: %lld  tc: %lld\n", (long long)*pos, (long long)tc);
 
 	if (tc >= end)
 		*pos = start;
@@ -68,17 +75,44 @@ static void ntfs_cluster_update_zone_pos(ntfs_volume *vol, u8 zone, LCN tc)
 {
 	ntfs_log_trace("tc = %lld, zone = %d\n", (long long)tc, zone);
 	
-	if (zone == 1)
+	if (zone == ZONE_MFT)
 		ntfs_cluster_set_zone_pos(vol->mft_lcn, vol->mft_zone_end,
 					  &vol->mft_zone_pos, tc);
-	else if (zone == 2)
+	else if (zone == ZONE_DATA1)
 		ntfs_cluster_set_zone_pos(vol->mft_zone_end, vol->nr_clusters,
 					  &vol->data1_zone_pos, tc);
-	else /* zone == 4 */
+	else /* zone == ZONE_DATA2 */
 		ntfs_cluster_set_zone_pos(0, vol->mft_zone_start, 
 					  &vol->data2_zone_pos, tc);
 }
 
+/*
+ *		Unmark full zones when a cluster has been freed in a full zone
+ *
+ *	Next allocation will reuse the freed cluster
+ */
+
+static void update_full_status(ntfs_volume *vol, LCN lcn)
+{
+	if (lcn >= vol->mft_zone_end) {
+		if (vol->full_zones & ZONE_DATA1) {
+			ntfs_cluster_update_zone_pos(vol, ZONE_DATA1, lcn);
+			vol->full_zones &= ~ZONE_DATA1;
+		}
+	} else
+		if (lcn < vol->mft_zone_start) {
+			if (vol->full_zones & ZONE_DATA2) {
+				ntfs_cluster_update_zone_pos(vol, ZONE_DATA2, lcn);
+				vol->full_zones &= ~ZONE_DATA2;
+			}
+		} else {
+			if (vol->full_zones & ZONE_MFT) {
+				ntfs_cluster_update_zone_pos(vol, ZONE_MFT, lcn);
+				vol->full_zones &= ~ZONE_MFT;
+			}
+		}
+}
+ 
 static s64 max_empty_bit_range(unsigned char *buf, int size)
 {
 	int i, j, run = 0;
@@ -87,30 +121,49 @@ static s64 max_empty_bit_range(unsigned char *buf, int size)
 	
 	ntfs_log_trace("Entering\n");
 	
-	for (i = 0; i < size; i++, buf++) {
-		
-		if (*buf == 0) {
-			run += 8;
-			continue;
-		}
-		
-		for (j = 0; j < 8; j++) {
+	i = 0;
+	while (i < size) {
+		switch (*buf) {
+		case 0 :
+			do {
+				buf++;
+				run += 8;
+				i++;
+			} while ((i < size) && !*buf);
+			break;
+		case 255 :
+			if (run > max_range) {
+				max_range = run;
+				start_pos = (s64)i * 8 - run;
+			}
+			run = 0;
+			do {
+				buf++;
+				i++;
+			} while ((i < size) && (*buf == 255));
+			break;
+		default :
+			for (j = 0; j < 8; j++) {
 			
-			int bit = *buf & (1 << j);
+				int bit = *buf & (1 << j);
 		
-			if (bit) {
-				if (run > max_range) {
-					max_range = run;
-					start_pos = i * 8 + j - run;
-				}
-				run = 0;
-			} else 
-				run++;
-		}		
+				if (bit) {
+					if (run > max_range) {
+						max_range = run;
+						start_pos = (s64)i * 8 + (j - run);
+					}
+					run = 0;
+				} else 
+					run++;
+			}
+			i++;
+			buf++;
+		
+		}
 	}
 	
 	if (run > max_range)
-		start_pos = i * 8 - run;
+		start_pos = (s64)i * 8 - run;
 	
 	return start_pos;
 }
@@ -195,16 +248,17 @@ runlist *ntfs_cluster_alloc(ntfs_volume *vol, VCN start_vcn, s64 count,
 	u8 has_guess, used_zone_pos;
 	int err = 0, rlpos, rlsize, buf_size;
 
-	ntfs_log_trace("Entering with count = 0x%llx, start_lcn = 0x%llx, "
+	ntfs_log_enter("Entering with count = 0x%llx, start_lcn = 0x%llx, "
 		       "zone = %s_ZONE.\n", (long long)count, (long long)
 		       start_lcn, zone == MFT_ZONE ? "MFT" : "DATA");
+	
 	if (!vol || count < 0 || start_lcn < -1 || !vol->lcnbmp_na ||
 			(s8)zone < FIRST_ZONE || zone > LAST_ZONE) {
 		errno = EINVAL;
 		ntfs_log_perror("%s: vcn: %lld, count: %lld, lcn: %lld", 
 				__FUNCTION__, (long long)start_vcn, 
 				(long long)count, (long long)start_lcn);
-		return NULL;
+		goto out;
 	}
 
 	/* Return empty runlist if @count == 0 */
@@ -215,12 +269,12 @@ runlist *ntfs_cluster_alloc(ntfs_volume *vol, VCN start_vcn, s64 count,
 			rl[0].lcn = LCN_RL_NOT_MAPPED;
 			rl[0].length = 0;
 		}
-		return rl;
+		goto out;
 	}
 
 	buf = ntfs_malloc(NTFS_LCNALLOC_BSIZE);
 	if (!buf)
-		return NULL;
+		goto out;
 	/*
 	 * If no @start_lcn was requested, use the current zone
 	 * position otherwise use the requested @start_lcn.
@@ -244,13 +298,13 @@ runlist *ntfs_cluster_alloc(ntfs_volume *vol, VCN start_vcn, s64 count,
 		
 	if (zone_start < vol->mft_zone_start) {
 		zone_end = vol->mft_zone_start;
-		search_zone = 4;
+		search_zone = ZONE_DATA2;
 	} else if (zone_start < vol->mft_zone_end) {
 		zone_end = vol->mft_zone_end;
-		search_zone = 1;
+		search_zone = ZONE_MFT;
 	} else {
 		zone_end = vol->nr_clusters;
-		search_zone = 2;
+		search_zone = ZONE_DATA1;
 	}
 	
 	bmp_pos = zone_start;
@@ -259,6 +313,9 @@ runlist *ntfs_cluster_alloc(ntfs_volume *vol, VCN start_vcn, s64 count,
 	clusters = count;
 	rlpos = rlsize = 0;
 	while (1) {
+			/* check whether we have exhausted the current zone */
+		if (search_zone & vol->full_zones)
+			goto zone_pass_done;
 		last_read_pos = bmp_pos >> 3;
 		br = ntfs_attr_pread(vol->lcnbmp_na, last_read_pos, 
 				     NTFS_LCNALLOC_BSIZE, buf);
@@ -311,31 +368,46 @@ runlist *ntfs_cluster_alloc(ntfs_volume *vol, VCN start_vcn, s64 count,
 			/* Allocate the bitmap bit. */
 			*byte |= bit;
 			writeback = 1;
-			if (vol->free_clusters <= 0) 
-				ntfs_log_error("Non-positive free clusters "
-					       "(%lld)!\n",
+			if (NVolFreeSpaceKnown(vol)) {
+				if (vol->free_clusters <= 0)
+					ntfs_log_error("Non-positive free"
+					       " clusters (%lld)!\n",
 						(long long)vol->free_clusters);
-			else	
-				vol->free_clusters--; 
+				else	
+					vol->free_clusters--;
+			}
 			
 			/*
 			 * Coalesce with previous run if adjacent LCNs.
 			 * Otherwise, append a new run.
 			 */
-			if (prev_lcn == lcn + bmp_pos - prev_run_len && rlpos)
+			if (prev_lcn == lcn + bmp_pos - prev_run_len && rlpos) {
+				ntfs_log_debug("Cluster coalesce: prev_lcn: "
+					       "%lld  lcn: %lld  bmp_pos: %lld  "
+					       "prev_run_len: %lld\n", 
+					       (long long)prev_lcn, 
+					       (long long)lcn, (long long)bmp_pos, 
+					       (long long)prev_run_len);
 				rl[rlpos - 1].length = ++prev_run_len;
-			else {
+			} else {
 				if (rlpos)
 					rl[rlpos].vcn = rl[rlpos - 1].vcn +
 							prev_run_len;
-				else
+				else {
 					rl[rlpos].vcn = start_vcn;
+					ntfs_log_debug("Start_vcn: %lld\n", 
+						       (long long)start_vcn);
+				}
 				
 				rl[rlpos].lcn = prev_lcn = lcn + bmp_pos;
 				rl[rlpos].length = prev_run_len = 1;
 				rlpos++;
 			}
 			
+			ntfs_log_debug("RUN:   %-16lld %-16lld %-16lld\n", 
+				       (long long)rl[rlpos - 1].vcn, 
+				       (long long)rl[rlpos - 1].lcn, 
+				       (long long)rl[rlpos - 1].length);
 			/* Done? */
 			if (!--clusters) {
 				if (used_zone_pos)
@@ -357,9 +429,9 @@ runlist *ntfs_cluster_alloc(ntfs_volume *vol, VCN start_vcn, s64 count,
 			
 			used_zone_pos = 1;
 			
-			if (search_zone == 1)
+			if (search_zone == ZONE_MFT)
 				zone_start = vol->mft_zone_pos;
-			else if (search_zone == 2)
+			else if (search_zone == ZONE_DATA1)
 				zone_start = vol->data1_zone_pos;
 			else
 				zone_start = vol->data2_zone_pos;
@@ -377,13 +449,12 @@ runlist *ntfs_cluster_alloc(ntfs_volume *vol, VCN start_vcn, s64 count,
 zone_pass_done:
 		ntfs_log_trace("Finished current zone pass(%i).\n", pass);
 		if (pass == 1) {
-			
 			pass = 2;
 			zone_end = zone_start;
 			
-			if (search_zone == 1)
+			if (search_zone == ZONE_MFT)
 				zone_start = vol->mft_zone_start;
-			else if (search_zone == 2)
+			else if (search_zone == ZONE_DATA1)
 				zone_start = vol->mft_zone_end;
 			else
 				zone_start = 0;
@@ -399,11 +470,12 @@ zone_pass_done:
 		/* pass == 2 */
 done_zones_check:
 		done_zones |= search_zone;
-		if (done_zones < 7) {
+		vol->full_zones |= search_zone;
+		if (done_zones < (ZONE_MFT + ZONE_DATA1 + ZONE_DATA2)) {
 			ntfs_log_trace("Switching zone.\n");
 			pass = 1;
 			if (rlpos) {
-				LCN tc = tc = rl[rlpos - 1].lcn + 
+				LCN tc = rl[rlpos - 1].lcn + 
 				      rl[rlpos - 1].length + NTFS_LCNALLOC_SKIP;
 				
 				if (used_zone_pos)
@@ -412,32 +484,32 @@ done_zones_check:
 			}
 			
 			switch (search_zone) {
-			case 1:
+			case ZONE_MFT:
 				ntfs_log_trace("Zone switch: mft -> data1\n");
-switch_to_data1_zone:		search_zone = 2;
+switch_to_data1_zone:		search_zone = ZONE_DATA1;
 				zone_start = vol->data1_zone_pos;
 				zone_end = vol->nr_clusters;
 				if (zone_start == vol->mft_zone_end)
 					pass = 2;
 				break;
-			case 2:
+			case ZONE_DATA1:
 				ntfs_log_trace("Zone switch: data1 -> data2\n");
-				search_zone = 4;
+				search_zone = ZONE_DATA2;
 				zone_start = vol->data2_zone_pos;
 				zone_end = vol->mft_zone_start;
 				if (!zone_start)
 					pass = 2;
 				break;
-			case 4:
-				if (!(done_zones & 2)) {
+			case ZONE_DATA2:
+				if (!(done_zones & ZONE_DATA1)) {
 					ntfs_log_trace("data2 -> data1\n");
 					goto switch_to_data1_zone;
 				}
 				ntfs_log_trace("Zone switch: data2 -> mft\n");
-				search_zone = 1;
+				search_zone = ZONE_MFT;
 				zone_start = vol->mft_zone_pos;
 				zone_end = vol->mft_zone_end;
-				if (!zone_start == vol->mft_zone_start)
+				if (zone_start == vol->mft_zone_start)
 					pass = 2;
 				break;
 			}
@@ -467,13 +539,15 @@ done_ret:
 		goto err_ret;
 	}
 done_err_ret:
-	ntfs_log_debug("At done_err_ret (follows done_ret).\n");
 	free(buf);
-	if (!err)
-		return rl;
-	ntfs_log_trace("Failed to allocate clusters (%d)", errno);
-	errno = err;
-	return NULL;
+	if (err) {
+		errno = err;
+		ntfs_log_perror("Failed to allocate clusters");
+		rl = NULL;
+	}
+out:	
+	ntfs_log_leave("\n");
+	return rl;
 
 wb_err_ret:
 	ntfs_log_trace("At wb_err_ret.\n");
@@ -486,6 +560,7 @@ err_ret:
 		rl[rlpos].vcn = rl[rlpos - 1].vcn + rl[rlpos - 1].length;
 		rl[rlpos].lcn = LCN_RL_NOT_MAPPED;
 		rl[rlpos].length = 0;
+		ntfs_debug_runlist_dump(rl);
 		ntfs_cluster_free_from_rl(vol, rl);
 		free(rl);
 		rl = NULL;
@@ -513,6 +588,7 @@ int ntfs_cluster_free_from_rl(ntfs_volume *vol, runlist *rl)
 			       (long long)rl->lcn, (long long)rl->length);
 
 		if (rl->lcn >= 0) { 
+			update_full_status(vol,rl->lcn);
 			if (ntfs_bitmap_clear_run(vol->lcnbmp_na, rl->lcn, 
 						  rl->length)) {
 				ntfs_log_perror("Cluster deallocation failed "
@@ -528,6 +604,43 @@ int ntfs_cluster_free_from_rl(ntfs_volume *vol, runlist *rl)
 	ret = 0;
 out:
 	vol->free_clusters += nr_freed; 
+	if (NVolFreeSpaceKnown(vol)
+	    && (vol->free_clusters > vol->nr_clusters))
+		ntfs_log_error("Too many free clusters (%lld > %lld)!",
+			       (long long)vol->free_clusters, 
+			       (long long)vol->nr_clusters);
+	return ret;
+}
+
+/*
+ *		Basic cluster run free
+ *	Returns 0 if successful
+ */
+
+int ntfs_cluster_free_basic(ntfs_volume *vol, s64 lcn, s64 count)
+{
+	s64 nr_freed = 0;
+	int ret = -1;
+
+	ntfs_log_trace("Entering.\n");
+	ntfs_log_trace("Dealloc lcn 0x%llx, len 0x%llx.\n",
+			       (long long)lcn, (long long)count);
+
+	if (lcn >= 0) { 
+		update_full_status(vol,lcn);
+		if (ntfs_bitmap_clear_run(vol->lcnbmp_na, lcn, 
+						  count)) {
+			ntfs_log_perror("Cluster deallocation failed "
+				       "(%lld, %lld)",
+					(long long)lcn, 
+					(long long)count);
+				goto out;
+		}
+		nr_freed += count; 
+	}
+	ret = 0;
+out:
+	vol->free_clusters += nr_freed;
 	if (vol->free_clusters > vol->nr_clusters)
 		ntfs_log_error("Too many free clusters (%lld > %lld)!",
 			       (long long)vol->free_clusters, 
@@ -563,22 +676,23 @@ int ntfs_cluster_free(ntfs_volume *vol, ntfs_attr *na, VCN start_vcn, s64 count)
 		errno = EINVAL;
 		return -1;
 	}
-	ntfs_log_trace("Entering for inode 0x%llx, attr 0x%x, count 0x%llx, "
+	
+	ntfs_log_enter("Entering for inode 0x%llx, attr 0x%x, count 0x%llx, "
 		       "vcn 0x%llx.\n", (unsigned long long)na->ni->mft_no,
-		       na->type, (long long)count, (long long)start_vcn);
+		       le32_to_cpu(na->type), (long long)count, (long long)start_vcn);
 
 	rl = ntfs_attr_find_vcn(na, start_vcn);
 	if (!rl) {
 		if (errno == ENOENT)
-			return 0;
-		return -1;
+			ret = 0;
+		goto leave;
 	}
 
 	if (rl->lcn < 0 && rl->lcn != LCN_HOLE) {
 		errno = EIO;
 		ntfs_log_perror("%s: Unexpected lcn (%lld)", __FUNCTION__, 
 				(long long)rl->lcn);
-		return -1;
+		goto leave;
 	}
 
 	/* Find the starting cluster inside the run that needs freeing. */
@@ -591,9 +705,10 @@ int ntfs_cluster_free(ntfs_volume *vol, ntfs_attr *na, VCN start_vcn, s64 count)
 
 	if (rl->lcn != LCN_HOLE) {
 		/* Do the actual freeing of the clusters in this run. */
+		update_full_status(vol,rl->lcn + delta);
 		if (ntfs_bitmap_clear_run(vol->lcnbmp_na, rl->lcn + delta,
 					  to_free))
-			return -1;
+			goto leave;
 		nr_freed = to_free;
 	} 
 
@@ -623,6 +738,7 @@ int ntfs_cluster_free(ntfs_volume *vol, ntfs_attr *na, VCN start_vcn, s64 count)
 			to_free = count;
 
 		if (rl->lcn != LCN_HOLE) {
+			update_full_status(vol,rl->lcn);
 			if (ntfs_bitmap_clear_run(vol->lcnbmp_na, rl->lcn,
 					to_free)) {
 				// FIXME: Eeek! We need rollback! (AIA)
@@ -652,5 +768,7 @@ out:
 		ntfs_log_error("Too many free clusters (%lld > %lld)!",
 			       (long long)vol->free_clusters, 
 			       (long long)vol->nr_clusters);
+leave:	
+	ntfs_log_leave("\n");
 	return ret;
 }
