@@ -5,6 +5,7 @@
  * Copyright (c) 2004-2005 Richard Russon
  * Copyright (c) 2004-2008 Szabolcs Szakacsits
  * Copyright (c)      2005 Yura Pakhuchiy
+ * Copyright (c) 2014-2021 Jean-Pierre Andre
  *
  * This program/include file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as published
@@ -38,6 +39,9 @@
 #ifdef HAVE_STRING_H
 #include <string.h>
 #endif
+#ifdef HAVE_LIMITS_H
+#include <limits.h>
+#endif
 #include <time.h>
 
 #include "compat.h"
@@ -168,6 +172,15 @@ int ntfs_mft_records_write(const ntfs_volume *vol, const MFT_REF mref,
 		cnt = vol->mftmirr_size - m;
 		if (cnt > count)
 			cnt = count;
+		if ((m + cnt) > vol->mftmirr_na->initialized_size >>
+				vol->mft_record_size_bits) {
+			errno = ESPIPE;
+			ntfs_log_perror("Trying to write non-allocated mftmirr"
+				" records (%lld > %lld)", (long long)m + cnt,
+				(long long)vol->mftmirr_na->initialized_size >>
+				vol->mft_record_size_bits);
+			return -1;
+		}
 		bmirr = ntfs_malloc(cnt * vol->mft_record_size);
 		if (!bmirr)
 			return -1;
@@ -206,15 +219,32 @@ int ntfs_mft_records_write(const ntfs_volume *vol, const MFT_REF mref,
 	return -1;
 }
 
+/*
+ *		Check the consistency of an MFT record
+ *
+ *	Make sure its general fields are safe, then examine all its
+ *	attributes and apply generic checks to them.
+ *	The attribute checks are skipped when a record is being read in
+ *	order to collect its sequence number for creating a new record.
+ *
+ *	Returns 0 if the checks are successful
+ *		-1 with errno = EIO otherwise
+ */
+
 int ntfs_mft_record_check(const ntfs_volume *vol, const MFT_REF mref, 
 			  MFT_RECORD *m)
 {			  
 	ATTR_RECORD *a;
+	ATTR_TYPES previous_type;
 	int ret = -1;
+	u32 offset;
+	s32 space;
 	
 	if (!ntfs_is_file_record(m->magic)) {
-		ntfs_log_error("Record %llu has no FILE magic (0x%x)\n",
-			       (unsigned long long)MREF(mref), *(le32 *)m);
+		if (!NVolNoFixupWarn(vol))
+			ntfs_log_error("Record %llu has no FILE magic (0x%x)\n",
+				(unsigned long long)MREF(mref),
+				(int)le32_to_cpu(*(le32*)m));
 		goto err_out;
 	}
 	
@@ -225,13 +255,57 @@ int ntfs_mft_record_check(const ntfs_volume *vol, const MFT_REF mref,
 			       le32_to_cpu(m->bytes_allocated));
 		goto err_out;
 	}
-	
+	if (!NVolNoFixupWarn(vol)
+	    && (le32_to_cpu(m->bytes_in_use) > vol->mft_record_size)) {
+		ntfs_log_error("Record %llu has corrupt in-use size "
+			       "(%u > %u)\n", (unsigned long long)MREF(mref),
+			       (int)le32_to_cpu(m->bytes_in_use),
+			       (int)vol->mft_record_size);
+		goto err_out;
+	}
+	if (le16_to_cpu(m->attrs_offset) & 7) {
+		ntfs_log_error("Attributes badly aligned in record %llu\n",
+			       (unsigned long long)MREF(mref));
+		goto err_out;
+	}
+
 	a = (ATTR_RECORD *)((char *)m + le16_to_cpu(m->attrs_offset));
 	if (p2n(a) < p2n(m) || (char *)a > (char *)m + vol->mft_record_size) {
 		ntfs_log_error("Record %llu is corrupt\n",
 			       (unsigned long long)MREF(mref));
 		goto err_out;
 	}
+
+	if (!NVolNoFixupWarn(vol)) {
+		offset = le16_to_cpu(m->attrs_offset);
+		space = le32_to_cpu(m->bytes_in_use) - offset;
+		a = (ATTR_RECORD*)((char*)m + offset);
+		previous_type = AT_STANDARD_INFORMATION;
+		while ((space >= (s32)offsetof(ATTR_RECORD, resident_end))
+		    && (a->type != AT_END)
+		    && (le32_to_cpu(a->type) >= le32_to_cpu(previous_type))) {
+			if ((le32_to_cpu(a->length) <= (u32)space)
+			    && !(le32_to_cpu(a->length) & 7)) {
+				if (!ntfs_attr_inconsistent(a, mref)) {
+					previous_type = a->type;
+					offset += le32_to_cpu(a->length);
+					space -= le32_to_cpu(a->length);
+					a = (ATTR_RECORD*)((char*)m + offset);
+				} else
+					goto err_out;
+			} else {
+				ntfs_log_error("Corrupted MFT record %llu\n",
+				       (unsigned long long)MREF(mref));
+				goto err_out;
+			}
+		}
+			/* We are supposed to reach an AT_END */
+		if ((space < 4) || (a->type != AT_END)) {
+			ntfs_log_error("Bad end of MFT record %llu\n",
+				       (unsigned long long)MREF(mref));
+			goto err_out;
+		}
+	}
 	
 	ret = 0;
 err_out:
@@ -349,7 +423,7 @@ int ntfs_mft_record_layout(const ntfs_volume *vol, const MFT_REF mref,
 		 * Set the NTFS 3.1+ specific fields while we know that the
 		 * volume version is 3.1+.
 		 */
-		mrec->reserved = cpu_to_le16(0);
+		mrec->reserved = const_cpu_to_le16(0);
 		mrec->mft_record_number = cpu_to_le32(MREF(mref));
 	}
 	mrec->magic = magic_FILE;
@@ -357,7 +431,7 @@ int ntfs_mft_record_layout(const ntfs_volume *vol, const MFT_REF mref,
 		mrec->usa_count = cpu_to_le16(vol->mft_record_size /
 				NTFS_BLOCK_SIZE + 1);
 	else {
-		mrec->usa_count = cpu_to_le16(1);
+		mrec->usa_count = const_cpu_to_le16(1);
 		ntfs_log_error("Sector size is bigger than MFT record size.  "
 				"Setting usa_count to 1.  If Windows chkdsk "
 				"reports this as corruption, please email %s "
@@ -366,14 +440,14 @@ int ntfs_mft_record_layout(const ntfs_volume *vol, const MFT_REF mref,
 				"Thank you.\n", NTFS_DEV_LIST);
 	}
 	/* Set the update sequence number to 1. */
-	*(u16*)((u8*)mrec + le16_to_cpu(mrec->usa_ofs)) = cpu_to_le16(1);
-	mrec->lsn = cpu_to_le64(0ull);
-	mrec->sequence_number = cpu_to_le16(1);
-	mrec->link_count = cpu_to_le16(0);
+	*(le16*)((u8*)mrec + le16_to_cpu(mrec->usa_ofs)) = const_cpu_to_le16(1);
+	mrec->lsn = const_cpu_to_sle64(0ll);
+	mrec->sequence_number = const_cpu_to_le16(1);
+	mrec->link_count = const_cpu_to_le16(0);
 	/* Aligned to 8-byte boundary. */
 	mrec->attrs_offset = cpu_to_le16((le16_to_cpu(mrec->usa_ofs) +
 			(le16_to_cpu(mrec->usa_count) << 1) + 7) & ~7);
-	mrec->flags = cpu_to_le16(0);
+	mrec->flags = const_cpu_to_le16(0);
 	/*
 	 * Using attrs_offset plus eight bytes (for the termination attribute),
 	 * aligned to 8-byte boundary.
@@ -381,11 +455,11 @@ int ntfs_mft_record_layout(const ntfs_volume *vol, const MFT_REF mref,
 	mrec->bytes_in_use = cpu_to_le32((le16_to_cpu(mrec->attrs_offset) + 8 +
 			7) & ~7);
 	mrec->bytes_allocated = cpu_to_le32(vol->mft_record_size);
-	mrec->base_mft_record = cpu_to_le64((MFT_REF)0);
-	mrec->next_attr_instance = cpu_to_le16(0);
+	mrec->base_mft_record = const_cpu_to_le64((MFT_REF)0);
+	mrec->next_attr_instance = const_cpu_to_le16(0);
 	a = (ATTR_RECORD*)((u8*)mrec + le16_to_cpu(mrec->attrs_offset));
 	a->type = AT_END;
-	a->length = cpu_to_le32(0);
+	a->length = const_cpu_to_le32(0);
 	/* Finally, clear the unused part of the mft record. */
 	memset((u8*)a + 8, 0, vol->mft_record_size - ((u8*)a + 8 - (u8*)mrec));
 	return 0;
@@ -548,7 +622,7 @@ static int ntfs_mft_bitmap_find_free_rec(ntfs_volume *vol, ntfs_inode *base_ni)
 					"data_pos 0x%llx, bit 0x%llx, "
 					"*byte 0x%hhx, b %u.\n", size,
 					(long long)data_pos, (long long)bit,
-					byte ? *byte : -1, b);
+					(u8) (byte ? *byte : -1), b);
 			for (; bit < size && data_pos + bit < pass_end;
 					bit &= ~7ull, bit += 8) {
 				/* 
@@ -575,7 +649,7 @@ static int ntfs_mft_bitmap_find_free_rec(ntfs_volume *vol, ntfs_inode *base_ni)
 					"data_pos 0x%llx, bit 0x%llx, "
 					"*byte 0x%hhx, b %u.\n", size,
 					(long long)data_pos, (long long)bit,
-					byte ? *byte : -1, b);
+					(u8) (byte ? *byte : -1), b);
 			data_pos += size;
 			/*
 			 * If the end of the pass has not been reached yet,
@@ -718,7 +792,7 @@ static int ntfs_mft_bitmap_extend_allocation_i(ntfs_volume *vol)
 		goto undo_alloc;
 	}
 	/* Get the size for the new mapping pairs array for this extent. */
-	mp_size = ntfs_get_size_for_mapping_pairs(vol, rl2, ll);
+	mp_size = ntfs_get_size_for_mapping_pairs(vol, rl2, ll, INT_MAX);
 	if (mp_size <= 0) {
 		ntfs_log_error("Get size for mapping pairs failed for "
 				"mft bitmap attribute extent.\n");
@@ -1025,7 +1099,7 @@ static int ntfs_mft_data_extend_allocation(ntfs_volume *vol)
 				"count %lli.\n", (long long)nr);
 	} while (1);
 	
-	ntfs_log_debug("Allocated %lli clusters.\n", nr);
+	ntfs_log_debug("Allocated %lld clusters.\n", (long long)nr);
 	
 	rl = ntfs_runlists_merge(mft_na->rl, rl2);
 	if (!rl) {
@@ -1067,7 +1141,7 @@ static int ntfs_mft_data_extend_allocation(ntfs_volume *vol)
 		goto undo_alloc;
 	}
 	/* Get the size for the new mapping pairs array for this extent. */
-	mp_size = ntfs_get_size_for_mapping_pairs(vol, rl2, ll);
+	mp_size = ntfs_get_size_for_mapping_pairs(vol, rl2, ll, INT_MAX);
 	if (mp_size <= 0) {
 		ntfs_log_error("Get size for mapping pairs failed for "
 				"mft data attribute extent.\n");
@@ -1187,7 +1261,7 @@ undo_alloc:
 static int ntfs_mft_record_init(ntfs_volume *vol, s64 size)
 {
 	int ret = -1;
-	ntfs_attr *mft_na, *mftbmp_na;
+	ntfs_attr *mft_na;
 	s64 old_data_initialized, old_data_size;
 	ntfs_attr_search_ctx *ctx;
 	
@@ -1196,7 +1270,6 @@ static int ntfs_mft_record_init(ntfs_volume *vol, s64 size)
 	/* NOTE: Caller must sanity check vol, vol->mft_na and vol->mftbmp_na */
 	
 	mft_na = vol->mft_na;
-	mftbmp_na = vol->mftbmp_na;
 	
 	/*
 	 * The mft record is outside the initialized data. Extend the mft data
@@ -1292,14 +1365,13 @@ undo_data_init:
 static int ntfs_mft_rec_init(ntfs_volume *vol, s64 size)
 {
 	int ret = -1;
-	ntfs_attr *mft_na, *mftbmp_na;
+	ntfs_attr *mft_na;
 	s64 old_data_initialized, old_data_size;
 	ntfs_attr_search_ctx *ctx;
 	
 	ntfs_log_enter("Entering\n");
 	
 	mft_na = vol->mft_na;
-	mftbmp_na = vol->mftbmp_na;
 	
 	if (size > mft_na->allocated_size || size > mft_na->initialized_size) {
 		errno = EIO;
@@ -1351,7 +1423,7 @@ undo_data_init:
 	goto out;
 }
 
-static ntfs_inode *ntfs_mft_rec_alloc(ntfs_volume *vol)
+ntfs_inode *ntfs_mft_rec_alloc(ntfs_volume *vol, BOOL mft_data)
 {
 	s64 ll, bit;
 	ntfs_attr *mft_na, *mftbmp_na;
@@ -1359,7 +1431,8 @@ static ntfs_inode *ntfs_mft_rec_alloc(ntfs_volume *vol)
 	ntfs_inode *ni = NULL;
 	ntfs_inode *base_ni;
 	int err;
-	u16 seq_no, usn;
+	le16 seq_no, usn;
+	BOOL forced_mft_data;
 
 	ntfs_log_enter("Entering\n");
 
@@ -1368,7 +1441,49 @@ static ntfs_inode *ntfs_mft_rec_alloc(ntfs_volume *vol)
 
 	base_ni = mft_na->ni;
 
-	bit = ntfs_mft_bitmap_find_free_rec(vol, base_ni);
+	/*
+	 * The first extent containing $MFT:$AT_DATA is better located
+	 * in record 15 to make sure it can be read at mount time.
+	 * The record 15 is prereserved as a base inode with no
+	 * extents and no name, and it is marked in use.
+	 */
+	forced_mft_data = FALSE;
+	if (mft_data) {
+		ntfs_inode *ext_ni = ntfs_inode_open(vol, FILE_mft_data);
+			/*
+			 * If record 15 cannot be opened, it is probably in
+			 * use as an extent. Apply standard procedure for
+			 * further extents.
+			 */
+		if (ext_ni) {
+			/*
+			 * Make sure record 15 is a base extent and it has
+			 * no name. A base inode with no name cannot be in use.
+			 * The test based on base_mft_record fails for
+			 * extents of MFT, so we need a special check.
+			 * If already used, apply standard procedure.
+			 */
+   			if (!ext_ni->mrec->base_mft_record
+			    && !ext_ni->mrec->link_count)
+				forced_mft_data = TRUE;
+			ntfs_inode_close(ext_ni);
+			/* Double-check, in case it is used for MFT */
+			if (forced_mft_data && base_ni->nr_extents) {
+				int i;
+
+				for (i=0; i<base_ni->nr_extents; i++) {
+					if (base_ni->extent_nis[i]
+					    && (base_ni->extent_nis[i]->mft_no
+							== FILE_mft_data))
+						forced_mft_data = FALSE;
+   				}
+			}
+		}
+	}
+	if (forced_mft_data)
+		bit = FILE_mft_data;
+	else
+		bit = ntfs_mft_bitmap_find_free_rec(vol, base_ni);
 	if (bit >= 0)
 		goto found_free_rec;
 
@@ -1405,25 +1520,36 @@ found_free_rec:
 		goto undo_mftbmp_alloc;
 	}
 	/* Sanity check that the mft record is really not in use. */
-	if (ntfs_is_file_record(m->magic) && (m->flags & MFT_RECORD_IN_USE)) {
+	if (!forced_mft_data
+	    && (ntfs_is_file_record(m->magic)
+	    && (m->flags & MFT_RECORD_IN_USE))) {
 		ntfs_log_error("Inode %lld is used but it wasn't marked in "
 			       "$MFT bitmap. Fixed.\n", (long long)bit);
 		free(m);
 		goto undo_mftbmp_alloc;
 	}
 
+		/*
+		 * Retrieve the former seq_no and usn so that the new record
+		 * cannot be mistaken for the former one.
+		 * However the original record may just be garbage, so
+		 * use some sensible value when they cannot be retrieved.
+		 */
 	seq_no = m->sequence_number;
-	usn = *(u16*)((u8*)m + le16_to_cpu(m->usa_ofs));
+	if (le16_to_cpu(m->usa_ofs) <= (NTFS_BLOCK_SIZE - 2))
+		usn = *(le16*)((u8*)m + (le16_to_cpu(m->usa_ofs) & -2));
+	else
+		usn = const_cpu_to_le16(1);
 	if (ntfs_mft_record_layout(vol, bit, m)) {
 		ntfs_log_error("Failed to re-format mft record.\n");
 		free(m);
 		goto undo_mftbmp_alloc;
 	}
-	if (le16_to_cpu(seq_no))
+	if (seq_no)
 		m->sequence_number = seq_no;
-	seq_no = le16_to_cpu(usn);
-	if (seq_no && seq_no != 0xffff)
-		*(u16*)((u8*)m + le16_to_cpu(m->usa_ofs)) = usn;
+	seq_no = usn;
+	if (seq_no && seq_no != const_cpu_to_le16(0xffff))
+		*(le16*)((u8*)m + le16_to_cpu(m->usa_ofs)) = usn;
 	/* Set the mft record itself in use. */
 	m->flags |= MFT_RECORD_IN_USE;
 	/* Now need to open an ntfs inode for the mft record. */
@@ -1459,7 +1585,7 @@ found_free_rec:
 			free(ni);
 			goto undo_mftbmp_alloc;
 		}
-		if (base_ni->extent_nis) {
+		if (base_ni->nr_extents) {
 			memcpy(extent_nis, base_ni->extent_nis,
 					i - 4 * sizeof(ntfs_inode *));
 			free(base_ni->extent_nis);
@@ -1472,10 +1598,10 @@ found_free_rec:
 	ntfs_inode_mark_dirty(ni);
 	/* Initialize time, allocated and data size in ntfs_inode struct. */
 	ni->data_size = ni->allocated_size = 0;
-	ni->flags = 0;
+	ni->flags = const_cpu_to_le32(0);
 	ni->creation_time = ni->last_data_change_time =
 			ni->last_mft_change_time =
-			ni->last_access_time = time(NULL);
+			ni->last_access_time = ntfs_current_time();
 	/* Update the default mft allocation position if it was used. */
 	if (!base_ni)
 		vol->mft_data_pos = bit + 1;
@@ -1521,8 +1647,9 @@ err_out:
  * @base_ni is NULL we start where we last stopped and we perform wrap around
  * when we reach the end.  Note, we do not try to allocate mft records below
  * number 24 because numbers 0 to 15 are the defined system files anyway and 16
- * to 24 are special in that they are used for storing extension mft records
- * for the $DATA attribute of $MFT.  This is required to avoid the possibility
+ * to 24 are used for storing extension mft records or used by chkdsk to store
+ * its log. However the record number 15 is dedicated to the first extent to
+ * the $DATA attribute of $MFT.  This is required to avoid the possibility
  * of creating a run list with a circular dependence which once written to disk
  * can never be read in again.  Windows will only use records 16 to 24 for
  * normal files if the volume is completely out of space.  We never use them
@@ -1588,7 +1715,9 @@ ntfs_inode *ntfs_mft_record_alloc(ntfs_volume *vol, ntfs_inode *base_ni)
 	MFT_RECORD *m;
 	ntfs_inode *ni = NULL;
 	int err;
-	u16 seq_no, usn;
+	u32 usa_ofs;
+	le16 seq_no, usn;
+	BOOL oldwarn;
 
 	if (base_ni)
 		ntfs_log_enter("Entering (allocating an extent mft record for "
@@ -1602,7 +1731,7 @@ ntfs_inode *ntfs_mft_record_alloc(ntfs_volume *vol, ntfs_inode *base_ni)
 	}
 	
 	if (ntfs_is_mft(base_ni)) {
-		ni = ntfs_mft_rec_alloc(vol);
+		ni = ntfs_mft_rec_alloc(vol, FALSE);
 		goto out;
 	}
 
@@ -1702,10 +1831,22 @@ found_free_rec:
 	if (!m)
 		goto undo_mftbmp_alloc;
 	
+	/*
+	 * As this is allocating a new record, do not expect it to have
+	 * been initialized previously, so do not warn over bad fixups
+	 * (hence avoid warn flooding when an NTFS partition has been wiped).
+	 */
+	oldwarn = !NVolNoFixupWarn(vol);
+	NVolSetNoFixupWarn(vol);
 	if (ntfs_mft_record_read(vol, bit, m)) {
+		if (oldwarn)
+			NVolClearNoFixupWarn(vol);
 		free(m);
 		goto undo_mftbmp_alloc;
 	}
+	if (oldwarn)
+		NVolClearNoFixupWarn(vol);
+
 	/* Sanity check that the mft record is really not in use. */
 	if (ntfs_is_file_record(m->magic) && (m->flags & MFT_RECORD_IN_USE)) {
 		ntfs_log_error("Inode %lld is used but it wasn't marked in "
@@ -1714,17 +1855,26 @@ found_free_rec:
 		goto retry;
 	}
 	seq_no = m->sequence_number;
-	usn = *(u16*)((u8*)m + le16_to_cpu(m->usa_ofs));
+		/*
+		 * As ntfs_mft_record_read() returns what has been read
+		 * even when the fixups have been found bad, we have to
+		 * check where we fetch the initial usn from.
+		 */
+	usa_ofs = le16_to_cpu(m->usa_ofs);
+	if (!(usa_ofs & 1) && (usa_ofs < NTFS_BLOCK_SIZE)) {
+		usn = *(le16*)((u8*)m + usa_ofs);
+	} else
+		usn = const_cpu_to_le16(1);
 	if (ntfs_mft_record_layout(vol, bit, m)) {
 		ntfs_log_error("Failed to re-format mft record.\n");
 		free(m);
 		goto undo_mftbmp_alloc;
 	}
-	if (le16_to_cpu(seq_no))
+	if (seq_no)
 		m->sequence_number = seq_no;
-	seq_no = le16_to_cpu(usn);
-	if (seq_no && seq_no != 0xffff)
-		*(u16*)((u8*)m + le16_to_cpu(m->usa_ofs)) = usn;
+	seq_no = usn;
+	if (seq_no && seq_no != const_cpu_to_le16(0xffff))
+		*(le16*)((u8*)m + le16_to_cpu(m->usa_ofs)) = usn;
 	/* Set the mft record itself in use. */
 	m->flags |= MFT_RECORD_IN_USE;
 	/* Now need to open an ntfs inode for the mft record. */
@@ -1761,7 +1911,7 @@ found_free_rec:
 				free(ni);
 				goto undo_mftbmp_alloc;
 			}
-			if (base_ni->extent_nis) {
+			if (base_ni->nr_extents) {
 				memcpy(extent_nis, base_ni->extent_nis,
 						i - 4 * sizeof(ntfs_inode *));
 				free(base_ni->extent_nis);
@@ -1774,10 +1924,10 @@ found_free_rec:
 	ntfs_inode_mark_dirty(ni);
 	/* Initialize time, allocated and data size in ntfs_inode struct. */
 	ni->data_size = ni->allocated_size = 0;
-	ni->flags = 0;
+	ni->flags = const_cpu_to_le32(0);
 	ni->creation_time = ni->last_data_change_time =
 			ni->last_mft_change_time =
-			ni->last_access_time = time(NULL);
+			ni->last_access_time = ntfs_current_time();
 	/* Update the default mft allocation position if it was used. */
 	if (!base_ni)
 		vol->mft_data_pos = bit + 1;
@@ -1816,7 +1966,8 @@ int ntfs_mft_record_free(ntfs_volume *vol, ntfs_inode *ni)
 {
 	u64 mft_no;
 	int err;
-	u16 seq_no, old_seq_no;
+	u16 seq_no;
+	le16 old_seq_no;
 
 	ntfs_log_trace("Entering for inode 0x%llx.\n", (long long) ni->mft_no);
 
@@ -1856,7 +2007,11 @@ int ntfs_mft_record_free(ntfs_volume *vol, ntfs_inode *ni)
 	}
 
 	/* Throw away the now freed inode. */
+#if CACHE_NIDATA_SIZE
+	if (!ntfs_inode_real_close(ni)) {
+#else
 	if (!ntfs_inode_close(ni)) {
+#endif
 		vol->free_mft_records++; 
 		return 0;
 	}
@@ -1883,13 +2038,14 @@ sync_rollback:
  */
 int ntfs_mft_usn_dec(MFT_RECORD *mrec)
 {
-	u16 usn, *usnp;
+	u16 usn;
+	le16 *usnp;
 
 	if (!mrec) {
 		errno = EINVAL;
 		return -1;
 	}
-	usnp = (u16 *)((char *)mrec + le16_to_cpu(mrec->usa_ofs));
+	usnp = (le16*)((char*)mrec + le16_to_cpu(mrec->usa_ofs));
 	usn = le16_to_cpup(usnp);
 	if (usn-- <= 1)
 		usn = 0xfffe;
