@@ -3,7 +3,7 @@
  *
  *	This module is part of ntfs-3g library
  *
- * Copyright (c) 2008-2009 Jean-Pierre Andre
+ * Copyright (c) 2008-2021 Jean-Pierre Andre
  *
  * This program/include file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as published
@@ -37,17 +37,17 @@
 #ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
 #endif
-
-#ifdef HAVE_SETXATTR
-#include <sys/xattr.h>
+#ifdef MAJOR_IN_MKDEV
+#include <sys/mkdev.h>
 #endif
-
-#ifdef HAVE_SYS_SYSMACROS_H
+#ifdef MAJOR_IN_SYSMACROS
 #include <sys/sysmacros.h>
 #endif
 
+#include "compat.h"
 #include "types.h"
 #include "debug.h"
+#include "layout.h"
 #include "attrib.h"
 #include "inode.h"
 #include "dir.h"
@@ -58,18 +58,8 @@
 #include "logging.h"
 #include "misc.h"
 #include "reparse.h"
-
-/* the definitions in layout.h are wrong, we use names defined in
-  http://msdn.microsoft.com/en-us/library/aa365740(VS.85).aspx
-*/
-
-#define IO_REPARSE_TAG_DFS         const_cpu_to_le32(0x8000000A)
-#define IO_REPARSE_TAG_DFSR        const_cpu_to_le32(0x80000012)
-#define IO_REPARSE_TAG_HSM         const_cpu_to_le32(0xC0000004)
-#define IO_REPARSE_TAG_HSM2        const_cpu_to_le32(0x80000006)
-#define IO_REPARSE_TAG_MOUNT_POINT const_cpu_to_le32(0xA0000003)
-#define IO_REPARSE_TAG_SIS         const_cpu_to_le32(0x80000007)
-#define IO_REPARSE_TAG_SYMLINK     const_cpu_to_le32(0xA000000C)
+#include "xattrs.h"
+#include "ea.h"
 
 struct MOUNT_POINT_REPARSE_DATA {      /* reparse data for junctions */
 	le16	subst_name_offset;
@@ -88,6 +78,11 @@ struct SYMLINK_REPARSE_DATA {          /* reparse data for symlinks */
 	char	path_buffer[0];      /* above data assume this is char array */
 } ;
 
+struct WSL_LINK_REPARSE_DATA {
+	le32	type;
+	char	link[0];
+} ;
+
 struct REPARSE_INDEX {			/* index entry in $Extend/$Reparse */
 	INDEX_ENTRY_HEADER header;
 	REPARSE_INDEX_KEY key;
@@ -200,8 +195,13 @@ static u64 ntfs_fix_file_name(ntfs_inode *dir_ni, ntfschar *uname,
 				 */
 				lemref = entry->indexed_file;
 				mref = le64_to_cpu(lemref);
-				for (i=0; i<found->file_name_length; i++)
-					uname[i] = found->file_name[i];
+				if (NVolCaseSensitive(vol) || !vol->locase) {
+					for (i=0; i<found->file_name_length; i++)
+						uname[i] = found->file_name[i];
+				} else {
+					for (i=0; i<found->file_name_length; i++)
+						uname[i] = vol->locase[le16_to_cpu(found->file_name[i])];
+				}
 			}
 		}
 		ntfs_index_ctx_put(icx);
@@ -230,6 +230,17 @@ static char *search_absolute(ntfs_volume *vol, ntfschar *path,
 	ni = ntfs_inode_open(vol, (MFT_REF)FILE_root);
 	if (ni) {
 		start = 0;
+		/*
+		 * Examine and translate the path, until we reach either
+		 *  - the end,
+		 *  - an unknown item
+		 *  - a non-directory
+		 *  - another reparse point,
+		 * A reparse point is not dereferenced, it will be
+		 * examined later when the translated path is dereferenced,
+		 * however the final part of the path will not be adjusted
+		 * to correct case.
+		 */
 		do {
 			len = 0;
 			while (((start + len) < count)
@@ -247,9 +258,11 @@ static char *search_absolute(ntfs_volume *vol, ntfschar *path,
 			}
 		} while (ni
 		    && (ni->mrec->flags & MFT_RECORD_IS_DIRECTORY)
+		    && !(ni->flags & FILE_ATTR_REPARSE_POINT)
 		    && (start < count));
 	if (ni
-	    && (ni->mrec->flags & MFT_RECORD_IS_DIRECTORY ? isdir : !isdir))
+	    && ((ni->mrec->flags & MFT_RECORD_IS_DIRECTORY ? isdir : !isdir)
+		|| (ni->flags & FILE_ATTR_REPARSE_POINT)))
 		if (ntfs_ucstombs(path, count, &target, 0) < 0) {
 			if (target) {
 				free(target);
@@ -283,23 +296,36 @@ static char *search_relative(ntfs_inode *ni, ntfschar *path, int count)
 	int pos;
 	int lth;
 	BOOL ok;
+	BOOL morelinks;
 	int max = 32; /* safety */
 
 	pos = 0;
 	ok = TRUE;
+	morelinks = FALSE;
 	curni = ntfs_dir_parent_inode(ni);
-	while (curni && ok && (pos < (count - 1)) && --max) {
+		/*
+		 * Examine and translate the path, until we reach either
+		 *  - the end,
+		 *  - an unknown item
+		 *  - a non-directory
+		 *  - another reparse point,
+		 * A reparse point is not dereferenced, it will be
+		 * examined later when the translated path is dereferenced,
+		 * however the final part of the path will not be adjusted
+		 * to correct case.
+		 */
+	while (curni && ok && !morelinks && (pos < (count - 1)) && --max) {
 		if ((count >= (pos + 2))
 		    && (path[pos] == const_cpu_to_le16('.'))
 		    && (path[pos+1] == const_cpu_to_le16('\\'))) {
-			path[1] = const_cpu_to_le16('/');
+			path[pos+1] = const_cpu_to_le16('/');
 			pos += 2;
 		} else {
 			if ((count >= (pos + 3))
 			    && (path[pos] == const_cpu_to_le16('.'))
 			    &&(path[pos+1] == const_cpu_to_le16('.'))
 			    && (path[pos+2] == const_cpu_to_le16('\\'))) {
-				path[2] = const_cpu_to_le16('/');
+				path[pos+2] = const_cpu_to_le16('/');
 				pos += 3;
 				newni = ntfs_dir_parent_inode(curni);
 				if (curni != ni)
@@ -326,12 +352,18 @@ static char *search_relative(ntfs_inode *ni, ntfschar *path, int count)
 					if (!curni)
 						ok = FALSE;
 					else {
+						if (curni->flags & FILE_ATTR_REPARSE_POINT)
+							morelinks = TRUE;
 						if (ok && ((pos + lth) < count)) {
 							path[pos + lth] = const_cpu_to_le16('/');
 							pos += lth + 1;
+							if (morelinks
+							   && ntfs_inode_close(curni))
+								ok = FALSE;
 						} else {
 							pos += lth;
-							if ((ni->mrec->flags ^ curni->mrec->flags)
+							if (!morelinks
+							  && (ni->mrec->flags ^ curni->mrec->flags)
 							    & MFT_RECORD_IS_DIRECTORY)
 								ok = FALSE;
 							if (ntfs_inode_close(curni))
@@ -392,11 +424,42 @@ static int ntfs_drive_letter(ntfs_volume *vol, ntfschar letter)
 	return (ret);
 }
 
+/*
+ *		Check whether reparse data describes a valid wsl special file
+ *	which is either a socket, a fifo, or a character or block device 
+ *
+ *	Return zero if valid, otherwise returns a negative error code
+ */
+
+int ntfs_reparse_check_wsl(ntfs_inode *ni, const REPARSE_POINT *reparse)
+{
+	int res;
+
+	res = -EOPNOTSUPP;
+	switch (reparse->reparse_tag) {
+	case IO_REPARSE_TAG_AF_UNIX :
+	case IO_REPARSE_TAG_LX_FIFO :
+	case IO_REPARSE_TAG_LX_CHR :
+	case IO_REPARSE_TAG_LX_BLK :
+		if (!reparse->reparse_data_length
+		    && (ni->flags & FILE_ATTRIBUTE_RECALL_ON_OPEN))
+			res = 0;
+		break;
+	default :
+		break;
+	}
+	if (res)
+		errno = EOPNOTSUPP;
+	return (res);
+}
+
 /*
  *		Do some sanity checks on reparse data
  *
- *	The only general check is about the size (at least the tag must
- *	be present)
+ *	Microsoft reparse points have an 8-byte header whereas
+ *	non-Microsoft reparse points have a 24-byte header.  In each case,
+ *	'reparse_data_length' must equal the number of non-header bytes.
+ *
  *	If the reparse data looks like a junction point or symbolic
  *	link, more checks can be done.
  *
@@ -410,14 +473,23 @@ static BOOL valid_reparse_data(ntfs_inode *ni,
 	unsigned int lth;
 	const struct MOUNT_POINT_REPARSE_DATA *mount_point_data;
 	const struct SYMLINK_REPARSE_DATA *symlink_data;
+	const struct WSL_LINK_REPARSE_DATA *wsl_reparse_data;
 
 	ok = ni && reparse_attr
 		&& (size >= sizeof(REPARSE_POINT))
+		&& (reparse_attr->reparse_tag != IO_REPARSE_TAG_RESERVED_ZERO)
 		&& (((size_t)le16_to_cpu(reparse_attr->reparse_data_length)
-				 + sizeof(REPARSE_POINT)) == size);
+			 + sizeof(REPARSE_POINT)
+			 + ((reparse_attr->reparse_tag &
+			     IO_REPARSE_TAG_IS_MICROSOFT) ? 0 : sizeof(GUID))) == size);
 	if (ok) {
 		switch (reparse_attr->reparse_tag) {
 		case IO_REPARSE_TAG_MOUNT_POINT :
+			if (size < sizeof(REPARSE_POINT) +
+				   sizeof(struct MOUNT_POINT_REPARSE_DATA)) {
+				ok = FALSE;
+				break;
+			}
 			mount_point_data = (const struct MOUNT_POINT_REPARSE_DATA*)
 						reparse_attr->reparse_data;
 			offs = le16_to_cpu(mount_point_data->subst_name_offset);
@@ -430,6 +502,11 @@ static BOOL valid_reparse_data(ntfs_inode *ni,
 				ok = FALSE;
 			break;
 		case IO_REPARSE_TAG_SYMLINK :
+			if (size < sizeof(REPARSE_POINT) +
+				   sizeof(struct SYMLINK_REPARSE_DATA)) {
+				ok = FALSE;
+				break;
+			}
 			symlink_data = (const struct SYMLINK_REPARSE_DATA*)
 						reparse_attr->reparse_data;
 			offs = le16_to_cpu(symlink_data->subst_name_offset);
@@ -439,6 +516,22 @@ static BOOL valid_reparse_data(ntfs_inode *ni,
 				 + offs + lth)) > size)
 				ok = FALSE;
 			break;
+		case IO_REPARSE_TAG_LX_SYMLINK :
+			wsl_reparse_data = (const struct WSL_LINK_REPARSE_DATA*)
+						reparse_attr->reparse_data;
+			if ((le16_to_cpu(reparse_attr->reparse_data_length)
+					<= sizeof(wsl_reparse_data->type))
+			    || (wsl_reparse_data->type != const_cpu_to_le32(2)))
+				ok = FALSE;
+			break;
+		case IO_REPARSE_TAG_AF_UNIX :
+		case IO_REPARSE_TAG_LX_FIFO :
+		case IO_REPARSE_TAG_LX_CHR :
+		case IO_REPARSE_TAG_LX_BLK :
+			if (reparse_attr->reparse_data_length
+			    || !(ni->flags & FILE_ATTRIBUTE_RECALL_ON_OPEN))
+				ok = FALSE;
+			break;
 		default :
 			break;
 		}
@@ -566,8 +659,9 @@ static char *ntfs_get_fulllink(ntfs_volume *vol, ntfschar *junction,
  *		or NULL if there were some problem, as described by errno
  */
 
-static char *ntfs_get_abslink(ntfs_volume *vol, ntfschar *junction,
-			int count, const char *mnt_point, BOOL isdir)
+char *ntfs_get_abslink(ntfs_volume *vol, ntfschar *junction, int count,
+			const char *mnt_point __attribute__((unused)),
+			BOOL isdir)
 {
 	char *target;
 	char *fulltarget;
@@ -613,10 +707,11 @@ static char *ntfs_get_abslink(ntfs_volume *vol, ntfschar *junction,
 			target = search_absolute(vol, &junction[3],
 				count - 3, isdir);
 		if (target) {
-			fulltarget = (char*)ntfs_malloc(strlen(mnt_point)
+			fulltarget = (char*)ntfs_malloc(
+					strlen(vol->abs_mnt_point)
 					+ strlen(target) + 2);
 			if (fulltarget) {
-				strcpy(fulltarget,mnt_point);
+				strcpy(fulltarget,vol->abs_mnt_point);
 				strcat(fulltarget,"/");
 				strcat(fulltarget,target);
 			}
@@ -641,10 +736,11 @@ static char *ntfs_get_abslink(ntfs_volume *vol, ntfschar *junction,
 			    && (target[0] >= 'a')
 			    && (target[0] <= 'z'))
 				target[0] += 'A' - 'a';
-			fulltarget = (char*)ntfs_malloc(strlen(mnt_point)
+			fulltarget = (char*)ntfs_malloc(
+				strlen(vol->abs_mnt_point)
 				    + sizeof(mappingdir) + strlen(target) + 1);
 			if (fulltarget) {
-				strcpy(fulltarget,mnt_point);
+				strcpy(fulltarget,vol->abs_mnt_point);
 				strcat(fulltarget,"/");
 				strcat(fulltarget,mappingdir);
 				strcat(fulltarget,target);
@@ -686,8 +782,7 @@ static char *ntfs_get_rellink(ntfs_inode *ni, ntfschar *junction, int count)
  *			symbolic link or directory junction
  */
 
-char *ntfs_make_symlink(ntfs_inode *ni, const char *mnt_point,
-			int *pattr_size)
+char *ntfs_make_symlink(ntfs_inode *ni, const char *mnt_point)
 {
 	s64 attr_size = 0;
 	char *target;
@@ -697,6 +792,7 @@ char *ntfs_make_symlink(ntfs_inode *ni, const char *mnt_point,
 	REPARSE_POINT *reparse_attr;
 	struct MOUNT_POINT_REPARSE_DATA *mount_point_data;
 	struct SYMLINK_REPARSE_DATA *symlink_data;
+	struct WSL_LINK_REPARSE_DATA *wsl_link_data;
 	enum { FULL_TARGET, ABS_TARGET, REL_TARGET } kind;
 	ntfschar *p;
 	BOOL bad;
@@ -779,10 +875,25 @@ char *ntfs_make_symlink(ntfs_inode *ni, const char *mnt_point,
 				break;
 			}
 			break;
+		case IO_REPARSE_TAG_LX_SYMLINK :
+			wsl_link_data = (struct WSL_LINK_REPARSE_DATA*)
+						reparse_attr->reparse_data;
+			if (wsl_link_data->type == const_cpu_to_le32(2)) {
+				lth = le16_to_cpu(
+					reparse_attr->reparse_data_length)
+					- sizeof(wsl_link_data->type);
+				target = (char*)ntfs_malloc(lth + 1);
+				if (target) {
+					memcpy(target, wsl_link_data->link,
+						lth);
+					target[lth] = 0;
+					bad = FALSE;
+				}
+			}
+			break;
 		}
 		free(reparse_attr);
 	}
-	*pattr_size = attr_size;
 	if (bad)
 		errno = EOPNOTSUPP;
 	return (target);
@@ -809,6 +920,7 @@ BOOL ntfs_possible_symlink(ntfs_inode *ni)
 		switch (reparse_attr->reparse_tag) {
 		case IO_REPARSE_TAG_MOUNT_POINT :
 		case IO_REPARSE_TAG_SYMLINK :
+		case IO_REPARSE_TAG_LX_SYMLINK :
 			possible = TRUE;
 		default : ;
 		}
@@ -817,7 +929,6 @@ BOOL ntfs_possible_symlink(ntfs_inode *ni)
 	return (possible);
 }
 
-#ifdef HAVE_SETXATTR	/* extended attributes interface required */
 
 /*
  *			Set the index for new reparse data
@@ -856,7 +967,6 @@ static int set_reparse_index(ntfs_inode *ni, ntfs_index_context *xr,
 	return (ntfs_ie_add(xr,(INDEX_ENTRY*)&indx));
 }
 
-#endif /* HAVE_SETXATTR */
 
 /*
  *		Remove a reparse data index entry if attribute present
@@ -933,7 +1043,6 @@ static ntfs_index_context *open_reparse_index(ntfs_volume *vol)
 	return (xr);
 }
 
-#ifdef HAVE_SETXATTR	/* extended attributes interface required */
 
 /*
  *		Update the reparse data and index
@@ -999,7 +1108,6 @@ static int update_reparse_data(ntfs_inode *ni, ntfs_index_context *xr,
 	return (res);
 }
 
-#endif /* HAVE_SETXATTR */
 
 /*
  *		Delete a reparse index entry
@@ -1038,7 +1146,6 @@ int ntfs_delete_reparse_index(ntfs_inode *ni)
 	return (res);
 }
 
-#ifdef HAVE_SETXATTR	/* extended attributes interface required */
 
 /*
  *		Get the ntfs reparse data into an extended attribute
@@ -1090,6 +1197,11 @@ int ntfs_set_ntfs_reparse_data(ntfs_inode *ni,
 	ntfs_index_context *xr;
 
 	res = 0;
+			/*
+			 * reparse data compatibily with EA is not checked
+			 * any more, it is required by Windows 10, but may
+			 * lead to problems with earlier versions.
+			 */
 	if (ni && valid_reparse_data(ni, (const REPARSE_POINT*)value, size)) {
 		xr = open_reparse_index(ni->vol);
 		if (xr) {
@@ -1219,4 +1331,117 @@ int ntfs_remove_ntfs_reparse_data(ntfs_inode *ni)
 	return (res ? -1 : 0);
 }
 
-#endif /* HAVE_SETXATTR */
+/*
+ *		Set reparse data for a WSL type symlink
+ */
+
+int ntfs_reparse_set_wsl_symlink(ntfs_inode *ni,
+			const ntfschar *target, int target_len)
+{
+	int res;
+	int len;
+	int reparse_len;
+	char *utarget;
+	REPARSE_POINT *reparse;
+	struct WSL_LINK_REPARSE_DATA *data;
+
+	res = -1;
+	utarget = (char*)NULL;
+	len = ntfs_ucstombs(target, target_len, &utarget, 0);
+	if (len > 0) {
+		reparse_len = sizeof(REPARSE_POINT) + sizeof(data->type) + len;
+		reparse = (REPARSE_POINT*)malloc(reparse_len);
+		if (reparse) {
+			data = (struct WSL_LINK_REPARSE_DATA*)
+					reparse->reparse_data;
+			reparse->reparse_tag = IO_REPARSE_TAG_LX_SYMLINK;
+			reparse->reparse_data_length
+				= cpu_to_le16(sizeof(data->type) + len);
+			reparse->reserved = const_cpu_to_le16(0);
+			data->type = const_cpu_to_le32(2);
+			memcpy(data->link, utarget, len);
+			res = ntfs_set_ntfs_reparse_data(ni,
+				(char*)reparse, reparse_len, 0);
+			free(reparse);
+		}
+	}
+	free(utarget);
+	return (res);
+}
+
+/*
+ *		Set reparse data for a WSL special file other than a symlink
+ *	(socket, fifo, character or block device)
+ */
+
+int ntfs_reparse_set_wsl_not_symlink(ntfs_inode *ni, mode_t mode)
+{
+	int res;
+	int len;
+	int reparse_len;
+	le32 reparse_tag;
+	REPARSE_POINT *reparse;
+
+	res = -1;
+	len = 0;
+	switch (mode) {
+	case S_IFSOCK :
+		reparse_tag = IO_REPARSE_TAG_AF_UNIX;
+		break;
+	case S_IFIFO :
+		reparse_tag = IO_REPARSE_TAG_LX_FIFO;
+		break;
+	case S_IFCHR :
+		reparse_tag = IO_REPARSE_TAG_LX_CHR;
+		break;
+	case S_IFBLK :
+		reparse_tag = IO_REPARSE_TAG_LX_BLK;
+		break;
+	default :
+		len = -1;
+		errno = EOPNOTSUPP;
+		break;
+	}
+	if (len >= 0) {
+		reparse_len = sizeof(REPARSE_POINT) + len;
+		reparse = (REPARSE_POINT*)malloc(reparse_len);
+		if (reparse) {
+			reparse->reparse_tag = reparse_tag;
+			reparse->reparse_data_length = cpu_to_le16(len);
+			reparse->reserved = const_cpu_to_le16(0);
+			res = ntfs_set_ntfs_reparse_data(ni,
+				(char*)reparse, reparse_len, 0);
+			free(reparse);
+		}
+	}
+	return (res);
+}
+
+
+/*
+ *		Get the reparse data into a buffer
+ *
+ *	Returns the buffer if the reparse data exists and is valid
+ *		NULL otherwise (with errno set according to the cause).
+ *	When a buffer is returned, it has to be freed by caller.
+ */
+
+REPARSE_POINT *ntfs_get_reparse_point(ntfs_inode *ni)
+{
+	s64 attr_size = 0;
+	REPARSE_POINT *reparse_attr;
+
+	reparse_attr = (REPARSE_POINT*)NULL;
+	if (ni) {
+		reparse_attr = (REPARSE_POINT*)ntfs_attr_readall(ni,
+			AT_REPARSE_POINT,(ntfschar*)NULL, 0, &attr_size);
+		if (reparse_attr
+		    && !valid_reparse_data(ni, reparse_attr, attr_size)) {
+			free(reparse_attr);
+			reparse_attr = (REPARSE_POINT*)NULL;
+			errno = EINVAL;
+		}
+	} else
+		errno = EINVAL;
+	return (reparse_attr);
+}
